import {
  filter,
  map,
  withLatestFrom,
  distinctUntilChanged,
} from 'rxjs/operators';
import { OnInit, ViewChild } from '@angular/core';
import { Observable } from 'rxjs';
import { select } from '@ngrx/store';
import { Component } from '@angular/core';
import { MatSelectChange } from '@angular/material';
import { MatDialog } from '@angular/material';
import { Store } from '@ngrx/store';
import { MediaMatcher } from '@angular/cdk/layout';
import { MatSnackBar } from '@angular/material';
import { GlobalState } from '@app/core/state';
import { DegreePlan } from '@app/core/models/degree-plan';
import { Year } from '@app/core/models/year';
import { FormBuilder, FormGroup } from '@angular/forms';
import { DegreePlannerApiService } from '@app/degree-planner/services/api.service';
import * as selectors from '@app/degree-planner/store/selectors';
import * as utils from '@app/degree-planner/shared/utils';
import { Course, SubjectMapping } from '@app/core/models/course';

// Actions
import {
  SwitchPlan,
  MakePlanPrimary,
  CreatePlan,
  ChangePlanName,
  DeletePlan,
} from '@app/degree-planner/store/actions/plan.actions';

import {
  ModifyPlanDialogComponent,
  DialogMode,
} from './dialogs/modify-plan-dialog/modify-plan-dialog.component';
import { PromptDialogComponent } from '@app/shared/dialogs/prompt-dialog/prompt-dialog.component';
import {
  ToggleAcademicYear,
  CloseCourseSearch,
} from './store/actions/ui.actions';

@Component({
  selector: 'cse-degree-planner',
  templateUrl: './degree-planner.component.html',
  styleUrls: ['./degree-planner.component.scss'],
})
export class DegreePlannerComponent implements OnInit {
  public termsByAcademicYear: Object;
  public mobileView: MediaQueryList;
  public coursesData$: any;

  public degreePlan$: Observable<DegreePlan | undefined>;
  public allDegreePlans$: Observable<DegreePlan[]>;
  public firstActiveTermCode$: Observable<string | undefined>;
  public termsByYear$: Observable<Year[]>;
  public isCourseSearchOpen$: Observable<boolean>;
  public subjects$: Observable<object>;
  public activeTerms$: Observable<string[]>;
  public yearCodes$: Observable<string[]>;
  public dropZoneIds$: Observable<string[]>;

  // Search variables
  public queriedCourses: Course[];
  public hasResults: boolean;
  public isLoading: boolean;
  public courseSearchForm: FormGroup;
  public isCourseSearchVisible$: Observable<boolean>;

  public subjects: SubjectMapping;
  public filteredSubjects: SubjectMapping;

  constructor(
    private store: Store<GlobalState>,
    public mediaMatcher: MediaMatcher,
    public dialog: MatDialog,
    private snackBar: MatSnackBar,
    private fb: FormBuilder,
    private api: DegreePlannerApiService,
  ) {
    this.mobileView = mediaMatcher.matchMedia('(max-width: 900px)');
  }

  public ngOnInit() {
    this.degreePlan$ = this.store.pipe(
      select(selectors.selectVisibleDegreePlan),
      filter(isntUndefined),
    );

    this.allDegreePlans$ = this.store.pipe(
      select(selectors.selectAllDegreePlans),
    );

    this.isCourseSearchVisible$ = this.store.pipe(
      select(selectors.selectCourseSearchVisibility),
    );

    // State attributes needed to create the search form
    this.subjects$ = this.store.pipe(select(selectors.getSubjects));
    this.activeTerms$ = this.store.pipe(select(selectors.getActiveTerms));

    // Internal values used to manage loading state
    this.queriedCourses = [];
    this.hasResults = false;
    this.isLoading = false;

    this.yearCodes$ = this.store.pipe(
      select(selectors.selectAllVisibleYears),
      map(years => Object.keys(years)),
      distinctUntilChanged(utils.compareStringArrays),
    );

    this.store
      .pipe(select(selectors.getSubjectDescriptions))
      .subscribe(subjects => {
        this.subjects = { [-1]: 'All', ...subjects };
        this.filteredSubjects = { [-1]: 'All', ...subjects };
      });

    this.dropZoneIds$ = this.store.pipe(
      select(selectors.selectAllVisibleYears),
      utils.yearsToDropZoneIds(),
      distinctUntilChanged(utils.compareStringArrays),
    );

    // Deafults for the search form
    this.courseSearchForm = this.fb.group({
      term: '0000',
      subject: 'All',
      search: '',
    });

    this.courseSearchForm.valueChanges.subscribe(values => {
      if (values.subject === '') {
        this.filteredSubjects = this.subjects;
        return;
      }

      const filtered = {};

      // Filter the terms based on users search
      Object.entries(this.subjects).map(subject => {
        const [key, name] = subject;

        const search = name.replace(/\s/g, '');

        if (search.toLowerCase().indexOf(values.subject.toLowerCase()) === 0) {
          filtered[key] = name;
        }
      });

      this.filteredSubjects = filtered;
    });

    this.isCourseSearchOpen$ = this.store.pipe(
      select(selectors.isCourseSearchOpen),
    );
  }

  public handleAcademicYearToggle(year: Year): void {
    // this.store.dispatch(
    // new ToggleAcademicYear({ year: year.twoDigitYearCode.toString() }),
    // );
  }

  public handleDegreePlanChange(event: MatSelectChange): void {
    if (typeof event.value === 'number') {
      this.store.dispatch(new SwitchPlan({ newVisibleRoadmapId: event.value }));
    }
  }

  public onCreatePlanClick() {
    const data: DialogMode = { mode: 'create' };
    this.dialog
      .open(ModifyPlanDialogComponent, { data })
      .afterClosed()
      .subscribe((result: { name: string } | undefined) => {
        if (result !== undefined && typeof result.name === 'string') {
          const name = result.name;
          const action = new CreatePlan({ name, primary: false });
          this.store.dispatch(action);
        }
      });
  }

  public onRenamePlanClick(currentPlan: DegreePlan) {
    // const data: DialogMode = { mode: 'rename', oldName: currentPlan.name };
    this.dialog
      .open(PromptDialogComponent, {
        data: {
          initalValue: currentPlan.name,
          title: 'Rename Degree Plan',
          confirmText: 'Save',
          inputName: 'Plan name',
        },
      })
      .afterClosed()
      .subscribe((result: { confirmed: boolean; value: string }) => {
        const { confirmed, value } = result;

        if (confirmed) {
          const { roadmapId } = currentPlan;
          const action = new ChangePlanName({ roadmapId, newName: value });
          this.store.dispatch(action);
        }
      });
  }

  public onMakePrimayClick(_currentPlan: DegreePlan) {
    const data: DialogMode = { mode: 'makePrimary' };
    this.dialog
      .open(ModifyPlanDialogComponent, { data })
      .afterClosed()
      .subscribe((result: { areYouSure: true } | undefined) => {
        if (result !== undefined && result.areYouSure === true) {
          const action = new MakePlanPrimary();
          this.store.dispatch(action);
        }
      });
  }

  public onDeletePlanClick(currentPlan: DegreePlan) {
    if (currentPlan.primary) {
      this.snackBar.open('The primary degree plan cannot be deleted');
      return;
    }

    const data: DialogMode = { mode: 'delete', name: currentPlan.name };
    this.dialog
      .open(ModifyPlanDialogComponent, { data })
      .afterClosed()
      .pipe(withLatestFrom(this.store))
      .subscribe(([result, state]) => {
        if (typeof result === 'object' && result.areYouSure === true) {
          const { roadmapId } = currentPlan;
          const deleteAction = new DeletePlan({ roadmapId });
          this.store.dispatch(deleteAction);

          const primaryPlan = state.degreePlanner.allDegreePlans.find(plan => {
            return plan.primary;
          }) as DegreePlan;
          const newVisibleRoadmapId = primaryPlan.roadmapId;
          const switchPlanAction = new SwitchPlan({ newVisibleRoadmapId });
          this.store.dispatch(switchPlanAction);
        }
      });
  }

  public onPrintPlanClick() {
    // TODO
    console.log('onPrintPlanClick');
  }

  public onSharePlanClick() {
    // TODO
    console.log('onSharePlanClick');
  }

  public getTermDropZone() {
    const termCodes = ['favoriteCourse-dropZone'];

    for (const yearCode in this.termsByAcademicYear) {
      if (this.termsByAcademicYear[yearCode]) {
        const year = this.termsByAcademicYear[yearCode];
        for (const termKey in year.terms) {
          if (year.terms[termKey]) {
            const term = year.terms[termKey];
            termCodes.push('term-' + term.termCode);
          }
        }
      }
    }
    // console.log(termCodes);

    return termCodes;
  }

  public closeCourseSearch() {
    this.store.dispatch(new CloseCourseSearch());
  }

  public search() {
    // Get the form field values
    const { search, term, subject } = this.courseSearchForm.value;
    let subjectCode;

    // Check if subject is valid
    Object.entries(this.subjects).forEach(option => {
      const [key, value] = option;
      if (value === subject) {
        subjectCode = key;
      }
    });

    if (!subjectCode) {
      this.snackBar.open('Please select a valid subject', undefined, {
        duration: 6000,
      });
      return;
    }

    console.log(search, term, subject, subjectCode);

    // Set the internal UI state
    this.isLoading = true;
    this.hasResults = false;
    this.queriedCourses = [];

    // Hit the search API
    this.api
      .searchCourses({
        subjectCode,
        searchText: search,
        termCode: term === '' ? '0000' : term,
      })
      .toPromise()
      .then(res => {
        // TODO add error handeling

        // Update the internal state
        this.hasResults = true;
        this.isLoading = false;

        // Map out the results and update the course object to match the needed structure
        this.queriedCourses = res.hits.map(course => {
          return {
            ...course,
            subject: course.subject.description,
            subjectCode: course.subject.subjectCode,
          };
        });
      })
      .catch(console.log);
  }

  public resetSearch() {
    // Reset the internal state and form values
    this.queriedCourses = [];
    this.hasResults = false;
    this.courseSearchForm.setValue({
      subject: 'All',
      search: '',
      term: '0000',
    });
  }
}

const isntUndefined = <T>(anything: T | undefined): anything is T => {
  return anything !== undefined;
};
